import { useRouter } from "next/router";
import Image from "next/image";
import dynamic from "next/dynamic";

import {
  ActionIcon,
  Grid,
  MediaQuery,
  Space,
  useMantineTheme,
} from "@mantine/core";

import { Icon, ICONS } from "vseth-canine-ui";

import parse from "html-react-parser";

const NoAdblockBanner = dynamic(() => import("../components/noAdblockBanner"), {
  ssr: false,
});
import { getAccentColor } from "../utilities/colors";

import content from "../content/texts.json";
import tux from "../public/images/tux.png";
import logo from "../public/images/logo.svg";

export default function Header() {
  const { locale } = useRouter();
  const theme = useMantineTheme();

  return (
    <div
      style={{
        height: "calc(100vh - 112px)",
        display: "flex",
        flexDirection: "column",
        justifyContent: "space-evenly",
      }}
    >
      <NoAdblockBanner />
      <Grid style={{ display: "flex", alignItems: "center" }}>
        <Grid.Col xs={12} md={6}>
          <div>
            <Image
              src={
                theme.colorScheme === "dark"
                  ? "https://static.vseth.ethz.ch/assets/vseth-0522-thealt/logo-inv.svg"
                  : logo
              }
              width={0}
              height={0}
              sizes="100vw"
              style={{ width: "30rem", maxWidth: "100%", height: "auto" }}
              alt="TheAlternative Logo  "
            />
            <Space h="xl" />
            <Space h="xl" />
            {content.topSection.map((entry, i) => (
              <p key={i}>{parse(entry[locale || "en"])}</p>
            ))}
          </div>
        </Grid.Col>
        <MediaQuery smallerThan="md" styles={{ display: "none" }}>
          <Grid.Col md={6}>
            {/*<Image
              src={tux}
              width={0}
              height={0}
              sizes="100vw"
              style={{ width: "100%", height: "auto" }}
              alt="Tux, the Linux Mascot"
            />*/}
            <img
              src="/images/tux.png"
              style={{
                width: "100%",
                height: "auto",
                filter: "drop-shadow(0px 0px 250px rgba(255, 255, 255, 0.4))",
              }}
              alt="Tux, the Linux Mascot"
            />
          </Grid.Col>
        </MediaQuery>
      </Grid>
      <a href="#events">
        <ActionIcon
          style={{ marginLeft: "auto", marginRight: "auto" }}
          size="3rem"
        >
          <Icon icon={ICONS.DOWN} color={getAccentColor(theme)} size="2rem" />
        </ActionIcon>
      </a>
    </div>
  );
}
