import NextAuth from "next-auth";
import KeycloakProvider from "next-auth/providers/keycloak";
import EmailProvider from "next-auth/providers/email";
import GoogleProvider from "next-auth/providers/google";
const jwt = require("jsonwebtoken");
import { PrismaAdapter } from "@auth/prisma-adapter";

import prisma from "../../../lib/prisma";

const prismaAdapter = PrismaAdapter(prisma);

const MyAdapter = {
  ...prismaAdapter,
  linkAccount: (account) => {
    account["not_before_policy"] = account["not-before-policy"];
    delete account["not-before-policy"];
    return prismaAdapter.linkAccount(account);
  },
};

export const authOptions = {
  theme: {
    colorScheme: "auto",
    brandColor: "#244471", // Hex color code
    logo: "/images/logo.svg", // Absolute URL to image
    buttonText: "#f28a20", // Hex color code
  },
  pages: {
    signIn: "/login",
    verifyRequest: "/checkMail",
  },
  adapter: MyAdapter,
  providers: [
    KeycloakProvider({
      clientId: process.env.SIP_AUTH_WEBSITE_CLIENT_ID,
      clientSecret: process.env.SIP_AUTH_WEBSITE_CLIENT_SECRET,
      issuer: process.env.SIP_AUTH_OIDC_ISSUER,
    }),
    EmailProvider({
      server:
        "smtp://" +
        process.env.MAILER_USERNAME +
        ":" +
        process.env.MAILER_PASSWORD +
        "@" +
        process.env.MAILER_HOST +
        ":" +
        process.env.MAILER_PORT,
      from: process.env.MAILER_NAME,
    }),
  ],
  callbacks: {
    async session({ session, token, user }) {
      const getToken = await prisma.account.findFirst({
        where: {
          userId: user.id,
        },
      });

      if (getToken) {
        const accessToken = getToken.access_token;
        const decoded = jwt.decode(accessToken, { complete: true });
        if (decoded) {
          session.info = decoded;
          session.user.id = user.id;
        }
      }

      return session;
    },
  },
};

export default NextAuth(authOptions);
