import { useState, useEffect } from "react";

import { useRouter } from "next/router";

import { useSession, signIn } from "next-auth/react";

import { useTranslation } from "next-i18next";
import { serverSideTranslations } from "next-i18next/serverSideTranslations";

import {
  Button,
  Card,
  Container,
  Divider,
  Space,
  TextInput,
  useMantineTheme,
} from "@mantine/core";
import { showNotification } from "@mantine/notifications";

import { Icon, ICONS } from "vseth-canine-ui";

import { getAccentColor } from "../utilities/colors";

export default function Login() {
  const theme = useMantineTheme();
  const { t } = useTranslation("common");

  const { data: session } = useSession();

  const { push } = useRouter();

  useEffect(() => {
    if (session) push("/");
  }, [session]);

  const [email, setEmail] = useState("");

  const keycloakSignIn = () => {
    signIn("keycloak");
  };

  const emailSignIn = (e) => {
    e.preventDefault();
    if (email == "") {
      showNotification({
        title: t("noEmail"),
        message: t("noEmailText"),
        color: "red",
      });
      return;
    }
    signIn("email", { email });
  };

  return (
    <Container size="xs">
      <Card withBorder shadow="md">
        <div style={{ textAlign: "center" }}>
          {theme.colorScheme === "dark" ? (
            <img
              src="https://static.vseth.ethz.ch/assets/vseth-0522-thealt/logo-inv.svg"
              alt="TheAlternative Logo"
              style={{ maxWidth: "100%", width: "20rem" }}
            />
          ) : (
            <img
              src="/images/logo.svg"
              alt="TheAlternative Logo"
              style={{ maxWidth: "100%", width: "20rem" }}
            />
          )}
        </div>
        <Space h="xl" />
        <Button size="lg" fullWidth onClick={keycloakSignIn}>
          {t("signInWithInstitution")}
        </Button>
        <Space h="xl" />
        <Space h="xl" />
        <Divider label={t("or").toUpperCase()} labelPosition="center" />
        <Space h="xl" />
        <form>
          <TextInput
            label="Email"
            placeholder={t("yourEmail")}
            value={email}
            onChange={(e) => setEmail(e.target.value)}
            icon={
              <Icon
                icon={ICONS.EMAIL}
                size={14}
                color={getAccentColor(theme)}
              />
            }
          />
          <Space h="xs" />
          <Button fullWidth onClick={(e) => emailSignIn(e)} type="submit">
            {t("signInWithEmail")}
          </Button>
        </form>
      </Card>
    </Container>
  );
}

export async function getStaticProps({ locale }) {
  return {
    props: {
      ...(await serverSideTranslations(locale, ["common"])),
      protected: false,
    },
  };
}
